(* file: lists.ml
   author: Bob Muller

   CS1103 Computer Science I Honors

   Lecture Notes and Code for Meeting 2 of Week 4

  Topics:
    1. More on Working with Lists
    2. A Greedy Algorithm for Making Change
*)

(* copy : 'a -> int -> 'a list
*)
let rec copy x n =
  match n = 0 with
  | true  -> []
  | false -> x :: copy x (n - 1)

(* swapAll : ('a * 'b) list -> ('b * 'a) list
*)
let rec swapAll pairs =
  match pairs with
  | [] -> []
  | (x, y) :: pairs -> (y, x) :: swapAll pairs

(* nth : 'a list -> int -> 'a

    The call (nth xs n) returns the nth element of xs.
    Note that the first element of xs is at index 0.
*)
let rec nth xs n =
  match n = 0 with
  | true ->
    (match xs with
     | [] -> failwith "nth: no such element"
     | x :: _ -> x)               (* NB: don't care variable for tl. *)
  | false -> nth (List.tl xs) (n - 1)

(* This version uses a pair to combine the two match-cases into one.
 *)
let rec nth xs n =
  match (n = 0, xs) with
  | (true, []) -> failwith "nth: no such element"
  | (true, x :: _) -> x
  | (false, xs) -> nth (List.tl xs) (n - 1)

(* range : int -> int list

   A call (range n) yields a list [0; 1; ...; n - 1].
*)
let range n =
  let rec range n =          (* NB: recycling function name range! *)
    match n = 0 with
    | true  -> []
    | false -> (n - 1) :: range (n - 1)
  in
  List.rev (range n)         (* NB: (range n) is backwards. *)

(* No need to use the reverse function (!) this saves a linear factor.
*)
let range n =
  let rec range n answer =
    match n = 0 with
    | true  -> answer
    | false -> let m = n - 1
               in
               range m (m :: answer)
  in
  range n []

(* Making Change w US Coinage --- a slightly larger problem.

    makeChange : int -> int list -> int list

  A call (makeChange amount coins) should return the shortest
  list of coins adding up to amount. For example, the call

    (makeChange 73 [1; 10; 5; 25])

  should return the list [25; 25; 10; 10; 1; 1; 1]. The Algorithm
  that we'll implement is an example of a "greedy algorithm" because
  the strategy --- pick the largest coin first --- sounds, well,
  greedy. The greedy algorithm will always find the minimal number of
  coins but only for certain coinange. For example, for the coinage
  [1; 10; 25] the greedy algorithm making change for 30 would produce
  [25; 1; 1; 1; 1; 1] but the shortest list would be three dimes
  [10; 10; 10].  A slightly more sophisticated algorithm can be
  used for the general case.

  First, a few helper functions.  Best read from the bottom up.

   remove : 'a -> 'a list -> 'a list
*)
let rec remove x xs =
  match xs with
  | [] -> []
  | y :: ys -> match x = y with
               | true  -> ys
               | false -> y :: remove x ys

(* maxList : int list -> int
*)
let rec maxList ns =
  match ns with
  | [] -> failwith "maxList: there are no numbers!"
  | [m] -> m
  | m :: ms -> max m (maxList ms)

(* getLargest : int list -> (int * int list)

   The call (getLargest coins) returns a pair (coin, coins') where
   coin is the largest coin in coins and coins' is coins with coin
   removed.
*)
let getLargest coins =
  let largest = maxList coins in
  let others  = remove largest coins
  in
  (largest, others)

(* makeChange : int -> int list -> int list
*)
let rec makeChange amount coins =
  match amount = 0 with
  | true  -> []
  | false ->
    let (largest, others) = getLargest coins in
    let (numLargest, residual) = (amount / largest, amount mod largest)
    in
    (copy largest numLargest) @ (makeChange residual others)
